
#toolbar{
 padding-bottom: 10px;
 }

#header-right li{
 list-style-type: none;
 }

.navbar a.navbar-brand {
    display: block;
    padding: 15px 20px;
}

div#header-center {
    padding: 0 0 0;
}

.navbar .nav > li > a {
   background: none;
}

div.facet div.facetTitle {
  background: none;
  font-weight: normal;
  border: none;
}

.facet{
 max-width: 80%;
 }

#focusLink{
 color: #FFFFFF;
}

.facetBody {
 border-style: none !important;
}

.facetBody .ui-slider{
 width: 80%;
 margin-left: 10%;
 }

.btn-facet{
 margin-bottom: 10px !important;
}

.navtoolbar{
 margin: 10px;
}

/* Out-context-view */
.ooview .well{
  padding: 5px;
  webkit-box-shadow: 0 5px 10px rgba(0, 0, 0, 0.2);
  -moz-box-shadow: 0 5px 10px rgba(0, 0, 0, 0.2);
  box-shadow: 0 5px 5px rgba(0, 0, 0, 0.2);
}

.ooview .glyphicon{
  font-size: 30px;
}

.ooview .well p.text-center{
  margin: 0 0 0px;
}


.admintools a{
 font-size: 12px;
 }

.contextFreeBox a{
 font-size: 12px;
 }

div.filter {
 background-color: white !important;
 border: 1px solid;
 }