.slickgrid {
  border: 1px solid #C0C0C0;
  border-top: none;
}

.slick-header-column.ui-state-default {
  height: 1.5em;
  line-height: 1.5em;
  padding: 0 4px;
}

.ui-widget-content .slick-cell a {
  color: #3399F3;
}
